package czerwony.kapturek.jblinehighlight;

import java.awt.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;

import java.beans.*;
import java.util.*;

import com.borland.primetime.ide.*;
import com.borland.jbuilder.viewer.*;
import com.borland.primetime.viewer.*;
import com.borland.jbuilder.node.*;
import com.borland.primetime.node.*;
import com.borland.primetime.editor.*;


/**
 * OpenTool for highlighting current line.
 * @todo layers - line outline should be on the top (not covered by selections)
 * @todo support for history windows
 * @todo support for split windows
 */
public class LineHighlightTool implements ChangeListener, PropertyChangeListener
{
    //private final static String TXT_CATEGORYNAME = "Curent line highlight";
    private final static String TXT_BLOCK = "Fill outline";
    private final static String TXT_ROUND = "Round corners";
    private final static String TXT_UNDERLINE = "Underline only";
    private final static String TXT_DASH = "Draw dashed line";
    private final static String TXT_HIDEONSELECT = "Hide while selecting text";
    private final static String TXT_COLOR = "Current line highlight";

    private final static Color  DEF_FOREGROUND = Color.black;
    private final static Color  DEF_BACKGROUND = Color.white;

    private static LineHighlightTool I = null;

    private EditorOptionCategory optionCategory;
    protected OutlineHighlightPainter painter;

    public static void initOpenTool( byte major, byte minor )
    {
        I = new LineHighlightTool();
    }

    public LineHighlightTool()
    {
        printWelcomeMessage();
        initConfigOptions();
        initPainter();
        Browser.addStaticBrowserListener(new ListenerImpl());
    }


    private void printWelcomeMessage()
    {
        System.out.println(
            "**********************************************************************\n" +
            "**  Initializing: jbLineHighlight (v0.3)                            **\n" +
            "**                by Czerwony Kapturek (czerwony.kapturek@wp.pl)    **\n" +
            "**********************************************************************"
        );
    }

    private void initConfigOptions()
    {
        Icon ico = new ImageIcon(getClass().getResource("currentline.gif"));
        optionCategory = new EditorOptionCategory(TXT_COLOR, ico);
        addBoolOption(TXT_BLOCK, TXT_BLOCK, false);
        addBoolOption(TXT_ROUND, TXT_ROUND, false);
        addBoolOption(TXT_UNDERLINE, TXT_UNDERLINE, false);
        addBoolOption(TXT_DASH, TXT_DASH, true);
        addBoolOption(TXT_HIDEONSELECT, TXT_HIDEONSELECT, false);

        MasterStyleContext styleCtx = EditorManager.getStyleContext();
        Style style = styleCtx.addStyle(TXT_COLOR, null);
        style.addAttribute(Style.NameAttribute, TXT_COLOR);
        StyleConstants.setBackground(style, DEF_BACKGROUND);
        StyleConstants.setForeground(style, DEF_FOREGROUND);
    }

    private void addBoolOption(String key, String name, boolean def)
    {
        EditorManager.registerEditorOption(new BooleanEditorOption(key, optionCategory, name, def));
    }

    private void initPainter()
    {
        //put defaults for colors and register listener for color changes
        MasterStyleContext styleCtx = EditorManager.getStyleContext();
        Style style = styleCtx.getStyle(TXT_COLOR);
        //put defaults for style, no listener required here
        Color f = StyleConstants.getForeground(style);
        Color b = StyleConstants.getBackground(style);
        painter = new OutlineHighlightPainter(f == null ? Color.black : f,
                                              b == null ? Color.white : b);

        style.addChangeListener(this);
        EditorManager.addPropertyChangeListener(TXT_DASH, this);
    }


    /**
     * Called when user changes the colors to be used for highlighting.
     */
    public void stateChanged(ChangeEvent e)
    {
        Style style = (Style)e.getSource();
        painter.setColors(StyleConstants.getForeground(style),
                          StyleConstants.getBackground(style));
    }


    /**
     * Style of the stroke changed
     */
    public void propertyChange(PropertyChangeEvent ev)
    {
        painter.prepareStroke();
    }


    ///////////////////////////////////////////////////////////////////////////
    //
    //  INNER CLASSES START HERE
    //

    public class ListenerImpl
        extends BrowserAdapter
        implements CaretListener, PropertyChangeListener
    {
        private final static String LINE_FRAME = "line-frame";

        int lineOffset = -1;
        int dotOffset = -1;
        boolean hide = true;

        public ListenerImpl()
        {
        }


        /**
         * Method that does the actual job of highlighting current line.
         */
        private void refreshHighlight(JTextComponent txt)
        {
            try
            {
                int dot = txt.getCaretPosition();
                Document doc = txt.getDocument();
                Element[] roots = doc.getRootElements();
                int lineIdx = roots[0].getElementIndex(dot);
                Element line = roots[0].getElement(lineIdx);

                int o1 = line.getStartOffset();
                int o2 = line.getEndOffset();
                boolean h = EditorManager.isBooleanOptionValue(TXT_HIDEONSELECT) && (dot != txt.getCaret().getMark());
                if ((lineOffset != o1)  //cursor moved vertically
                    || (dot - o1 < 2)    //typing in the beginning of the line:
                                        //  fix for problem of highlight dissapearing
                                        //  when typing in new line
                    || (dot >= o2 - 1) && (dot < dotOffset)
                                        //going backwards at the end of the line:
                                        //  fix for highlight moving one row down
                                        //  after pressing backspace key at the end of the line
                    || h != hide)       //text selection appeard or dismissed
                {
                    //save current values
                    lineOffset = line.getStartOffset();
                    dotOffset = dot;
                    hide = h;

//                    System.err.println("updating highlights");
                    Highlighter highlighter = txt.getHighlighter();
                    Object hi = txt.getClientProperty(LINE_FRAME);
                    if (hi != null)
                    {
                        highlighter.removeHighlight(hi);
                        txt.putClientProperty(LINE_FRAME, null);
                    }
                    if (!hide)
                    {
                        hi = highlighter.addHighlight(line.getStartOffset(), line.getEndOffset(), painter);
                        txt.putClientProperty(LINE_FRAME, hi);
                    }
                }
            }
            catch (BadLocationException ex)
            {
                ex.printStackTrace();
            }
        }

        ///////////////////////////////////////////////////////////////////////
        // INTERFACES

        /**
         * Called whem caret is moved: invokes refreshHighlight() method.
         * This event covers 99% of cases (exception: deleting line).
         */
        public void caretUpdate(CaretEvent e)
        {
            JTextComponent txt = (JTextComponent)e.getSource();
            refreshHighlight(txt);
        }


        /**
         * Called when new tab is activated. We have to register caret listener here.
         */
        public void browserNodeActivated(Browser b, Node n)
        {
            //only two nodes types are supported
            if(n != null && ((n instanceof TextFileNode) || (n instanceof ClassFileNode)))
            {
                lineOffset = -1;
                com.borland.primetime.ide.NodeViewer nodeViewers[] = b.getViewers(n);
                for(int i = 0; i < nodeViewers.length; i++)
                {
                    if(nodeViewers[i] instanceof AbstractTextNodeViewer)
                    {
                        AbstractTextNodeViewer viewer = (AbstractTextNodeViewer)nodeViewers[i];
                        JTextComponent txt = viewer.getEditor();
                        //easy part: add caret listener
                        txt.addCaretListener(this);
                        //ugly part: when a line is deleted there is no caret event issued
                        //registed listener for document changes
                        Document doc = txt.getDocument();
                        if (doc != null)
                            doc.addDocumentListener(new DocListener(txt));
                        //additional listener in case view changes the document
                        txt.addPropertyChangeListener("document", this);
                        //initial update: to highlight the first line of newly opened files
                        refreshHighlight(txt);
                    }
                }

            }
        }


        /**
         * Unregister all listeners for node/document.
         */
        public void browserNodeClosed(Browser b, Node n)
        {
            if(n != null && ((n instanceof TextFileNode) || (n instanceof ClassFileNode)))
            {
                com.borland.primetime.ide.NodeViewer nodeViewers[] = b.getViewers(n);
                for(int i = 0; i < nodeViewers.length; i++)
                {
                    if(nodeViewers[i] instanceof TextNodeViewer)
                    {
                        TextNodeViewer viewer = (TextNodeViewer)nodeViewers[i];
                        JTextComponent txt = viewer.getEditor();
                        txt.removeCaretListener(this);
                        txt.removePropertyChangeListener(this);
                        Document doc = txt.getDocument();
                        if (doc != null)
                        {
                            EventListener[] el = ((AbstractDocument)doc).getListeners(DocListener.class);
                            if (el.length > 0)
                                doc.removeDocumentListener((DocListener)el[0]);
                        }
                    }
                }

            }
        }


        /**
         * Document changed inside the component.
         * Remove listeners from ald document and use the new one.
         */
        public void propertyChange(PropertyChangeEvent ev)
        {
            JTextComponent comp = (JTextComponent)ev.getSource();
            AbstractDocument doc = (AbstractDocument)ev.getNewValue();
            doc.addDocumentListener(new DocListener(comp));

            doc = (AbstractDocument)ev.getOldValue();
            EventListener[] el = doc.getListeners(DocListener.class);
            if (el.length > 0)
                ((Document)ev.getOldValue()).removeDocumentListener((DocListener)el[0]);
        }


        /**
         * Document listener - needed when deleting text.
         */
        class DocListener implements DocumentListener
        {
            JTextComponent comp;

            DocListener(JTextComponent comp)
            {
                this.comp = comp;
            }

            public void insertUpdate(DocumentEvent e)
            {
            }

            public void removeUpdate(DocumentEvent e)
            {
                lineOffset = -1;
                refreshHighlight(comp);
            }

            public void changedUpdate(DocumentEvent e)
            {
            }
        }
    }


    /**
     * Class that implements the actual component responsible for painting
     * the line highlight.
     */
    class OutlineHighlightPainter extends LayeredHighlighter.LayerPainter
    {
        Color foreground;
        Color background;
        BasicStroke stroke = null;

        OutlineHighlightPainter(Color foreground, Color background)
        {
            setColors(foreground, background);
        }

        void setColors(Color foreground, Color background)
        {
            this.foreground = foreground;
            this.background = background;
        }


        private void prepareStroke()
        {
            if (EditorManager.isBooleanOptionValue(TXT_DASH))
            {
                float dash1[] = {1.0f};
                stroke = new BasicStroke(1.0f, BasicStroke.CAP_BUTT,
                              BasicStroke.JOIN_ROUND, 1.0f, dash1, 0.0f);
            }
            else
            {
                stroke = new BasicStroke(1.0f);
            }
        }


        private Rectangle calculateOutlineRect( int p0,
                                            int p1,
                                            Shape s,
                                            JTextComponent editor,
                                            View v)
        {
            try
            {
                Shape s0 = v.modelToView(p0, s, Position.Bias.Backward);
                int x = 0;
                int dx = 0;
                int y = s0.getBounds().y;
                int dy = s0.getBounds().height - 1; //<-with this looks better in last line
                if (editor.getParent() instanceof JViewport)
                {
                    x = ((JViewport)editor.getParent()).getViewPosition().x +1;
                    dx = ((JViewport)editor.getParent()).getExtentSize().width - 3;
                }
                else
                {
                    x = editor.getMargin().left - 1;
                    dx = editor.getWidth() - editor.getMargin().left - editor.getMargin().right-2;
                }
                return new Rectangle(x, y, dx, dy);
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
            }
            return null;
        }


        public Shape paintLayer(Graphics g,
                                int p0,
                                int p1,
                                Shape s,
                                JTextComponent editor,
                                View v)
        {
            Graphics2D g2 = (Graphics2D)g;
            if (stroke == null )
            {
                prepareStroke();
            }

            boolean round = EditorManager.isBooleanOptionValue(TXT_ROUND);
            boolean justUnderline = EditorManager.isBooleanOptionValue(TXT_UNDERLINE);
            Rectangle r = calculateOutlineRect(p0, p1, s, editor, v);
            if (r == null)
                return null;

            if (EditorManager.isBooleanOptionValue(TXT_BLOCK))
            {
                g.setColor(background);
                if (round)
                {
                    int c = r.height*3/4;
                    g.fillRoundRect(r.x, r.y, r.width, r.height+1, c, c);
                }
                else
                {
                    g.fillRect(r.x, r.y, r.width, r.height+1);
                }
            }
            Stroke oldStroke = g2.getStroke();
            g.setColor(foreground);
            g2.setStroke(stroke);
            if (justUnderline)
            {
                g.drawLine(r.x, r.y+r.height, r.x+r.width, r.y+r.height);
            }
            else if (round)
            {
                int c = r.height*3/4;
                g.drawRoundRect(r.x, r.y, r.width, r.height, c, c);
            }
            else
            {
                g.drawRect(r.x, r.y, r.width, r.height);
            }
            g2.setStroke(oldStroke);
            r.height += 1;
            return r;
        }


        public void paint(Graphics g, int offs0, int offs1, Shape s, JTextComponent c)
        {
        }
    }
}
